@@ -9,18 +9,20 @@
  * version 2 of the License, or (at your option) any later version.
  *
  * \note
- * This library is distributed in the hope that it will be useful, 
- * but WITHOUT ANY WARRANTY; without even the implied warranty of 
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details. 
+ * Lesser General Public License for more details.
  *
  * \note
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA  02110-1301  USA
  */
 
+#define _DEFAULT_SOURCE
+
 #include "config.h"
 #include <gphoto2/gphoto2-list.h>
 #include <gphoto2/gphoto2-port-log.h>
@@ -31,81 +33,24 @@
 
 #include <gphoto2/gphoto2-result.h>
 
-#define CHECK_NULL(r)        {if (!(r)) return (GP_ERROR_BAD_PARAMETERS);}
 #define CHECK_RESULT(result) {int r = (result); if (r < 0) return (r);}
 
-/** check that the list is valid */
-#define CHECK_LIST(list)				\
-  do {							\
-    if (NULL == list) {					\
-      return (GP_ERROR_BAD_PARAMETERS);			\
-    } else if (list->ref_count == 0) {			\
-      /* catch calls to already freed list */		\
-      return (GP_ERROR_BAD_PARAMETERS);			\
-    }							\
-  } while (0)
-
-
-/**
- * Whether to preserve binary compatibility for structure internals.
- *
- * Question: Find out whether any libgphoto2 frontend relies on those
- *           internals.
- * Answer: They do, at least by instantiation a "struct CameraList foo;".
- *
- * Binary compatibility and problems:
- *   * fixed string size for name and value
- *   * fixed max number of entries
- *   * some apps instantiate a "struct CameraList foo;" directly and
- *     thus reserve a specific amount of memory
- *   * FIXME: Do frontends rely on the memory layout and size of their 
- *     "struct CameraList" or "struct CameraList *" when accessing
- *     its members? They *SHOULD* use access functions.
- */
-#undef CAMERALIST_STRUCT_COMPATIBILITY
-#define CAMERALIST_STRUCT_COMPATIBILITY
-
-
 /**
  * Internal _CameraList data structure
  **/
 
-#ifdef CAMERALIST_STRUCT_COMPATIBILITY
-
-#define MAX_ENTRIES 1024
-#define MAX_LIST_STRING_LENGTH 128
-struct _CameraList {
-	int  count;
-	struct {
-		char name  [MAX_LIST_STRING_LENGTH];
-		char value [MAX_LIST_STRING_LENGTH];
-	} entry [MAX_ENTRIES];
-	int ref_count;
-};
-
-/* check that the given index is in the valid range for this list */
-#define CHECK_INDEX_RANGE(list, index)			\
-  do {							\
-    if (index < 0 || index >= list->count) {		\
-      return (GP_ERROR_BAD_PARAMETERS);			\
-    }							\
-  } while (0)
-
-#else /* !CAMERALIST_STRUCT_COMPATIBILITY */
-
-#error The !CAMERALIST_STRUCT_COMPATIBILITY part has not been implemented yet!
-
-struct _CamaraSubList {
-	int count;
-  foo bar;
+struct _entry {
+	char *name;
+	char *value;
 };
 
 struct _CameraList {
-	int ref_count;
+	int	used;	/* used entries */
+	int	max;	/* allocated entries */
+	struct _entry *entry;
+	int	ref_count;
 };
 
-#endif /* !CAMERALIST_STRUCT_COMPATIBILITY */
-
 
 /**
  * \brief Creates a new #CameraList.
@@ -117,12 +62,9 @@ struct _CameraList {
 int
 gp_list_new (CameraList **list)
 {
-	CHECK_NULL (list);
+	C_PARAMS (list);
 
-	(*list) = malloc (sizeof (CameraList));
-        if (!(*list))
-		return (GP_ERROR_NO_MEMORY);
-	memset ((*list), 0, sizeof (CameraList));
+	C_MEM (*list = calloc (1, sizeof (CameraList)));
 
 	(*list)->ref_count = 1;
 
@@ -139,7 +81,7 @@ gp_list_new (CameraList **list)
 int
 gp_list_ref (CameraList *list)
 {
-	CHECK_LIST (list);
+	C_PARAMS (list && list->ref_count);
 
 	list->ref_count++;
 
@@ -158,7 +100,7 @@ gp_list_ref (CameraList *list)
 int
 gp_list_unref (CameraList *list)
 {
-	CHECK_LIST (list);
+	C_PARAMS (list && list->ref_count);
 
 	if (list->ref_count == 1) /* time to free */
 		gp_list_free (list);
@@ -175,17 +117,23 @@ gp_list_unref (CameraList *list)
  *
  **/
 int
-gp_list_free (CameraList *list) 
+gp_list_free (CameraList *list)
 {
-	CHECK_LIST (list);
+	int	i;
+	C_PARAMS (list && list->ref_count);
 
+	for (i=0;i<list->used;i++) {
+		free (list->entry[i].name);
+		list->entry[i].name = NULL;
+		free (list->entry[i].value);
+		list->entry[i].value = NULL;
+	}
+	free (list->entry);
 	/* Mark this list as having been freed. That may help us
 	 * prevent access to already freed lists.
 	 */
 	list->ref_count = 0;
-	
 	free (list);
-
         return (GP_OK);
 }
 
@@ -199,10 +147,17 @@ gp_list_free (CameraList *list)
 int
 gp_list_reset (CameraList *list)
 {
-	CHECK_LIST (list);
-
-	list->count = 0;
+	int i;
+	C_PARAMS (list && list->ref_count);
 
+	for (i=0;i<list->used;i++) {
+		free (list->entry[i].name);
+		list->entry[i].name = NULL;
+		free (list->entry[i].value);
+		list->entry[i].value = NULL;
+	}
+	/* keeps -> entry allocated for reuse. */
+	list->used = 0;
 	return (GP_OK);
 }
 
@@ -218,69 +173,33 @@ gp_list_reset (CameraList *list)
 int
 gp_list_append (CameraList *list, const char *name, const char *value)
 {
-	CHECK_LIST (list);
+	C_PARAMS (list && list->ref_count);
 
-	if (list->count == MAX_ENTRIES) {
-		gp_log (GP_LOG_ERROR, "gphoto2-list", "gp_list_append: "
-		"Tried to add more than %d entries to the list, reporting error.",
-		MAX_ENTRIES);
-		return (GP_ERROR_FIXED_LIMIT_EXCEEDED);
+	if (list->used == list->max) {
+		C_MEM (list->entry = realloc(list->entry,(list->max+100)*sizeof(struct _entry)));
+		list->max += 100;
 	}
 
 	if (name) {
-		/* check that the 'name' value fits */
-		const size_t buf_len = sizeof (list->entry[list->count].name);
-		const size_t str_len = strlen (name);
-		if (str_len >= buf_len) {
-			gp_log (GP_LOG_ERROR, "gphoto2-list", 
-				"gp_list_append: "
-				"'name' value too long (%d >= %d)",
-				str_len, buf_len);
-			return (GP_ERROR_FIXED_LIMIT_EXCEEDED);
-		}
-		/* set the value */
-		strncpy (list->entry[list->count].name, name, buf_len);
-		list->entry[list->count].name[buf_len-1] = '\0';
+		C_MEM (list->entry[list->used].name = strdup (name));
+	} else {
+		list->entry[list->used].name = NULL;
 	}
 	if (value) {
-		/* check that the 'value' value fits */
-		const size_t buf_len = sizeof (list->entry[list->count].value);
-		const size_t str_len = strlen (value);
-		if (str_len >= buf_len) {
-			gp_log (GP_LOG_ERROR, "gphoto2-list", 
-				"gp_list_append: "
-				"'value' value too long (%d >= %d)",
-				str_len, buf_len);
-			return (GP_ERROR_FIXED_LIMIT_EXCEEDED);
-		}
-		/* set the value */
-		strncpy (list->entry[list->count].value, value, buf_len);
-		list->entry[list->count].value[buf_len-1] = '\0';
+		C_MEM (list->entry[list->used].value = strdup (value));
+	} else {
+		list->entry[list->used].value = NULL;
 	}
-
-        list->count++;
-
+        list->used++;
         return (GP_OK);
 }
 
-/**
-  * \internal
-  */
-static void
-exchange (CameraList *list, int x, int y)
-{
-	char name  [128];
-	char value [128];
+static int
+cmp_list (const void *a, const void *b) {
+        const struct _entry *ca = a;
+        const struct _entry *cb = b;
 
-	/* here we use memcpy to avoid unterminated strings *
-	 * stored in the list. 128 is hardcoded. use a constant *
-	 * for cleaness */
-	memcpy (name,  list->entry[x].name, 128);
-	memcpy (value, list->entry[x].value, 128);
-	memcpy (list->entry[x].name,  list->entry[y].name, 128);
-	memcpy (list->entry[x].value, list->entry[y].value, 128);
-	memcpy (list->entry[y].name,  name, 128);
-	memcpy (list->entry[y].value, value, 128);
+        return strcmp (ca->name, cb->name);
 }
 
 /**
@@ -293,23 +212,15 @@ exchange (CameraList *list, int x, int y)
 int
 gp_list_sort (CameraList *list)
 {
-	int x, y, z;
+	C_PARAMS (list && list->ref_count);
 
-	CHECK_LIST (list);
-
-	for (x = 0; x < list->count - 1; x++)
-		for (y = x + 1; y < list->count; y++) {
-			z = strcmp (list->entry[x].name, list->entry[y].name);
-			if (z > 0)
-				exchange (list, x, y);
-		}
-
-	return (GP_OK);
+	qsort (list->entry, list->used, sizeof(list->entry[0]), cmp_list);
+	return GP_OK;
 }
 
 /**
  * Counts the entries in the \c list.
- * 
+ *
  * \param list a #CameraList
  * \return a gphoto2 error code
  *
@@ -317,9 +228,9 @@ gp_list_sort (CameraList *list)
 int
 gp_list_count (CameraList *list)
 {
-	CHECK_LIST (list);
+	C_PARAMS (list && list->ref_count);
 
-        return (list->count);
+        return (list->used);
 }
 
 /**
@@ -338,13 +249,13 @@ int
 gp_list_find_by_name (CameraList *list, int *index, const char *name)
 {
 	int i;
-	CHECK_LIST (list);
-	CHECK_NULL (name);
+	C_PARAMS (list && list->ref_count);
+	C_PARAMS (name);
 
 	/* We search backwards because our only known user
 	 * camlibs/ptp2/library.c thinks this is faster
 	 */
-	for (i=list->count-1; i >= 0; i--) {
+	for (i=list->used-1; i >= 0; i--) {
 	  if (0==strcmp(list->entry[i].name, name)) {
 	    if (index) {
 	      (*index) = i;
@@ -368,9 +279,9 @@ gp_list_find_by_name (CameraList *list, int *index, const char *name)
 int
 gp_list_get_name (CameraList *list, int index, const char **name)
 {
-	CHECK_LIST (list);
-	CHECK_NULL (name);
-	CHECK_INDEX_RANGE (list, index);
+	C_PARAMS (list && list->ref_count);
+	C_PARAMS (name);
+	C_PARAMS (0 <= index && index < list->used);
 
 	*name = list->entry[index].name;
 
@@ -389,9 +300,9 @@ gp_list_get_name (CameraList *list, int index, const char **name)
 int
 gp_list_get_value (CameraList *list, int index, const char **value)
 {
-	CHECK_LIST (list);
-	CHECK_NULL (value);
-	CHECK_INDEX_RANGE (list, index);
+	C_PARAMS (list && list->ref_count);
+	C_PARAMS (value);
+	C_PARAMS (0 <= index && index < list->used);
 
 	*value = list->entry[index].value;
 
@@ -410,26 +321,14 @@ gp_list_get_value (CameraList *list, int index, const char **value)
 int
 gp_list_set_value (CameraList *list, int index, const char *value)
 {
-	CHECK_LIST (list);
-	CHECK_NULL (value);
-	CHECK_INDEX_RANGE (list, index);
-
-	do {
-		/* check that the value fits */
-		const size_t buf_len = sizeof (list->entry[index].value);
-		const size_t str_len = strlen (value);
-		if (str_len >= buf_len) {
-			gp_log (GP_LOG_ERROR, "gphoto2-list", 
-				"gp_list_append: "
-				"'value' value too long (%d >= %d)",
-				str_len, buf_len);
-			return (GP_ERROR_FIXED_LIMIT_EXCEEDED);
-		}
-	} while (0);
-
-	/* set the value */
-	strcpy (list->entry[index].value, value);
+	char *newval;
+	C_PARAMS (list && list->ref_count);
+	C_PARAMS (value);
+	C_PARAMS (0 <= index && index < list->used);
 
+	C_MEM (newval = strdup(value));
+	free (list->entry[index].value);
+	list->entry[index].value = newval;
 	return (GP_OK);
 }
 
@@ -445,26 +344,14 @@ gp_list_set_value (CameraList *list, int index, const char *value)
 int
 gp_list_set_name (CameraList *list, int index, const char *name)
 {
-	CHECK_LIST (list);
-	CHECK_NULL (name);
-	CHECK_INDEX_RANGE (list, index);
-
-	do {
-		/* check that the value fits */
-		const size_t buf_len = sizeof (list->entry[index].name);
-		const size_t str_len = strlen (name);
-		if (str_len >= buf_len) {
-			gp_log (GP_LOG_ERROR, "gphoto2-list", 
-				"gp_list_append: "
-				"'name' value too long (%d >= %d)",
-				str_len, buf_len);
-			return (GP_ERROR_FIXED_LIMIT_EXCEEDED);
-		}
-	} while (0);
-
-	/* set the value */
-	strcpy (list->entry[index].name, name);
+	char *newname;
+	C_PARAMS (list && list->ref_count);
+	C_PARAMS (name);
+	C_PARAMS (0 <= index && index < list->used);
 
+	C_MEM (newname = strdup(name));
+	free (list->entry[index].name);
+	list->entry[index].name = newname;
 	return (GP_OK);
 }
 
@@ -488,8 +375,8 @@ gp_list_populate  (CameraList *list, const char *format, int count)
 	int x;
 	char buf[1024];
 
-	CHECK_LIST (list);
-	CHECK_NULL (format);
+	C_PARAMS (list && list->ref_count);
+	C_PARAMS (format);
 
 	gp_list_reset (list);
 	for (x = 0; x < count; x++) {
